import base64
import logging
from datetime import datetime, timedelta
from typing import Optional

from aiohttp import ClientSession
from cryptography.hazmat._types import _PUBLIC_KEY_TYPES as KeyTypes
from cryptography.hazmat.backends.openssl.backend import backend
from cryptography.x509 import load_der_x509_certificate
from fastapi import HTTPException, status

log = logging.getLogger('fastapi_azure_auth')


class ProviderConfig:
    def __init__(self) -> None:
        self.tenant_id = '9b5ff18e-53c0-45a2-8bc2-9c0c8f60b2c6'  # For non-intility apps, you need to change this.
        self._config_timestamp: Optional[datetime] = None

        self.authorization_endpoint: str
        self.signing_keys: list[KeyTypes]
        self.token_endpoint: str
        self.end_session_endpoint: str
        self.issuer: str

    async def load_config(self) -> None:
        """
        Loads config from the Intility openid-config endpoint if it's over 24 hours old (or don't exist)
        """
        refresh_time = datetime.now() - timedelta(hours=24)
        if not self._config_timestamp or self._config_timestamp < refresh_time:
            try:
                log.debug('Loading Intility Azure ID Provider configuration.')
                await self._load_openid_config()
                self._config_timestamp = datetime.now()
            except Exception as error:
                log.exception('Unable to fetch openid-configuration from Azure AD. Error: %s', error)
                # We can't fetch an up to date openid-config, so authentication will not work.
                if self._config_timestamp:
                    raise HTTPException(
                        status_code=status.HTTP_401_UNAUTHORIZED,
                        detail='Connection to Azure AD is down. Unable to fetch provider configuration',
                        headers={'WWW-Authenticate': 'Bearer'},
                    )
                else:
                    raise RuntimeError(f'Unable to fetch provider information. {error}')

            log.info('fastapi-azure-auth loaded settings from Azure AD.')
            log.info('authorization endpoint: %s', self.authorization_endpoint)
            log.info('token endpoint:         %s', self.token_endpoint)
            log.info('end session endpoint:   %s', self.end_session_endpoint)
            log.info('issuer:                 %s', self.issuer)

    async def _load_openid_config(self) -> None:
        """
        Load openid config, fetch signing keys
        """
        config_url = f'https://login.microsoftonline.com/{self.tenant_id}/v2.0/.well-known/openid-configuration'

        log.info('Trying to get OpenID Connect config from %s', config_url)
        async with ClientSession() as session:
            # Fetch openid config
            async with session.get(config_url, timeout=10) as openid_response:
                openid_response.raise_for_status()
                openid_cfg = await openid_response.json()
                jwks_uri = openid_cfg['jwks_uri']
                # Fetch keys
                async with session.get(jwks_uri, timeout=10) as jwks_response:
                    jwks_response.raise_for_status()
                    keys = await jwks_response.json()
                    signing_certificates = [x['x5c'][0] for x in keys['keys'] if x.get('use', 'sig') == 'sig']
                    self._load_keys(signing_certificates)

        self.authorization_endpoint = openid_cfg['authorization_endpoint']
        self.token_endpoint = openid_cfg['token_endpoint']
        self.end_session_endpoint = openid_cfg['end_session_endpoint']
        self.issuer = openid_cfg['issuer']

    def _load_keys(self, certificates: list[str]) -> None:
        """
        Create certificates based on signing keys and store them
        """
        new_keys = []
        for cert in certificates:
            log.debug('Loading public key from certificate: %s', cert)
            cert_obj = load_der_x509_certificate(base64.b64decode(cert), backend)
            new_keys.append(cert_obj.public_key())
        self.signing_keys = new_keys


provider_config = ProviderConfig()



import logging
from typing import Dict, Optional

from fastapi import FastAPI, status
from fastapi.exceptions import HTTPException
from fastapi.security import OAuth2AuthorizationCodeBearer
from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTClaimsError, JWTError
from starlette.requests import Request

from fastapi_azure_auth.provider_config import provider_config

log = logging.getLogger('fastapi_azure_auth')


def invalid_auth(detail: str) -> HTTPException:
    """
    Raise a 401 unauthorized with given detail message.
    """
    return HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail=detail,
        headers={'WWW-Authenticate': 'Bearer'},
    )


class AzureAuthorizationCodeBearer(OAuth2AuthorizationCodeBearer):
    def __init__(
        self, app: FastAPI, app_client_id: str, scopes: Optional[Dict[str, str]] = None, allow_guest_users: bool = True
    ) -> None:
        """
        Initialize settings.

        :param app: Your FastAPI app.
        :param app_client_id: Client ID for this app (not your SPA)
        :param scopes: Scopes, these are the ones you've configured in Azure AD. Key is scope, value is a description.
            Example:
                {
                    f'api://{settings.APP_CLIENT_ID}/user_impersonation': 'user impersonation'
                }
        :param allow_guest_users: Guest users in the tenant can by default access this app,
                                  unless service principals are set up. This setting allow you to deny this behaviour.
        """
        self.app_client_id: str = app_client_id
        self.allow_guest_users: bool = allow_guest_users

        @app.on_event('startup')
        async def load_config() -> None:
            """
            Load config on startup.
            """
            await provider_config.load_config()  # pragma: no cover

        super().__init__(
            authorizationUrl=f'https://login.microsoftonline.com/{provider_config.tenant_id}/oauth2/v2.0/authorize',
            tokenUrl=f'https://login.microsoftonline.com/{provider_config.tenant_id}/oauth2/v2.0/token',
            scopes=scopes,
            description='`Leave client_secret blank`',
        )

    async def __call__(self, request: Request) -> dict:
        """
        Extends call to also validate the token
        """
        access_token = await super().__call__(request=request)
        # Load new config if old
        await provider_config.load_config()
        for index, key in enumerate(provider_config.signing_keys):
            try:
                # Set strict in case defaults change
                options = {
                    'verify_signature': True,
                    'verify_aud': True,
                    'verify_iat': True,
                    'verify_exp': True,
                    'verify_nbf': True,
                    'verify_iss': True,
                    'verify_sub': True,
                    'verify_jti': True,
                    'verify_at_hash': True,
                    'require_aud': True,
                    'require_iat': True,
                    'require_exp': True,
                    'require_nbf': True,
                    'require_iss': True,
                    'require_sub': True,
                    'require_jti': False,
                    'require_at_hash': False,
                    'leeway': 0,
                }
                # Validate token and return claims
                token = jwt.decode(
                    access_token,
                    key=key,
                    algorithms=['RS256'],
                    audience=f'api://{self.app_client_id}',
                    issuer=f'https://sts.windows.net/{provider_config.tenant_id}/',
                    options=options,
                )
                if not self.allow_guest_users and token['tid'] != provider_config.tenant_id:
                    raise invalid_auth(detail='Guest users not allowed')
                return token
            except HTTPException:
                raise
            except JWTClaimsError as error:
                log.info('Token contains invalid claims. %s', error)
                raise invalid_auth(detail='Token contains invalid claims')
            except ExpiredSignatureError as error:
                log.info('Token signature has expired. %s', error)
                raise invalid_auth(detail='Token signature has expired')
            except JWTError as error:
                if str(error) == 'Signature verification failed.' and index < len(provider_config.signing_keys) - 1:
                    continue
                log.warning('Invalid token. Error: %s', error, exc_info=True)
                raise invalid_auth(detail='Unable to validate token')
            except Exception as error:
                # Extra failsafe in case of a bug in a future version of the jwt library
                log.exception('Unable to process jwt token. Uncaught error: %s', error)
                raise invalid_auth(detail='Unable to process token')
        raise invalid_auth(detail='Unable to verify token, no signing keys found')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_azure_auth']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'fastapi>=0.68.0,<0.69.0',
 'python-jose[cryptography]>=3.3.0,<4.0.0']

setup_kwargs = {
    'name': 'fastapi-azure-auth',
    'version': '1.0.0',
    'description': 'Azure AD authentication for Intility',
    'long_description': '<h1 align="center">\n  <img src=".github/images/intility.png" width="124px"/><br/>\n  FastAPI-Azure-auth\n</h1>\n\n<p align="center">\n    <em>Azure AD Authentication for FastAPI apps made easy.</em>\n</p>\n<p align="center">\n    <a href="https://python.org">\n        <img src="https://img.shields.io/badge/python-v3.9+-blue.svg" alt="Python version">\n    </a>\n    <a href="https://fastapi.tiangolo.com/">\n        <img src="https://img.shields.io/badge/FastAPI-0.68.0+%20-blue.svg" alt="FastAPI Version">\n    </a>\n</p>\n<p align="center">\n    <a href="https://codecov.io/gh/intility/fastapi-azure-auth">\n        <img src="https://codecov.io/gh/intility/fastapi-azure-auth/branch/main/graph/badge.svg" alt="Codecov">\n    </a>\n    <a href="https://github.com/pre-commit/pre-commit">\n        <img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white" alt="Pre-commit">\n    </a>\n    <a href="https://github.com/psf/black">\n        <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Black">\n    </a>\n    <a href="http://mypy-lang.org">\n        <img src="http://www.mypy-lang.org/static/mypy_badge.svg" alt="mypy">\n    </a>\n    <a href="https://pycqa.github.io/isort/">\n        <img src="https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336" alt="isort">\n    </a>\n</p>\n\n\n----------------\n\n## 🚀 Description\n\n> FastAPI is a modern, fast (high-performance), web framework for building APIs with Python, based on standard Python type hints.  \n  \nAt Intility, FastAPI is a popular framework among its developers, \nwith customer-facing and internal services developed entirely on a FastAPI backend.  \nThis package enables our developers to create features without worrying about authentication and authorization.  \n\nAlso, [we\'re hiring!](https://intility.no/en/career/)\n\n## ⚡️ Quick start\n### Azure\nAzure docs will be available when create-fastapi-app is developed. In the meantime \nplease use the [.NET](https://create.intility.app/dotnet/setup/authorization) documentation.\n\n\n### FastAPI\n\n1. Install this library:\n```bash\npip install fastapi-azure-auth\n# or\npoetry add fastapi-azure-auth\n```\n\n2. Include `swagger_ui_oauth2_redirect_url` and `swagger_ui_init_oauth` in your FastAPI app initialization:\n\n```python\napp = FastAPI(\n    ...\n    swagger_ui_oauth2_redirect_url=\'/oauth2-redirect\',\n    swagger_ui_init_oauth={\n        \'usePkceWithAuthorizationCodeGrant\': True, \n        \'clientId\': settings.OPENAPI_CLIENT_ID  # SPA app with grants to your app\n    },\n)\n```\n\n3. Ensure you have CORS enabled for your local environment, such as `http://localhost:8000`. See [main.py](main.py) \nand the `BACKEND_CORS_ORIGINS` in [config.py](demoproj/core/config.py) \n\n4. Import and configure your Azure authentication:\n\n```python\nfrom fastapi_azure_auth.auth import AzureAuthorizationCodeBearer\n\nazure_scheme = AzureAuthorizationCodeBearer(\n    app=app,\n    app_client_id=settings.APP_CLIENT_ID,  # Web app\n    scopes={\n        f\'api://{settings.APP_CLIENT_ID}/user_impersonation\': \'User Impersonation\',\n    },\n)\n```\n\n5. Set your `intility_scheme` as a dependency for your wanted views/routers:\n\n```python\napp.include_router(api_router, prefix=settings.API_V1_STR, dependencies=[Depends(azure_scheme)])\n```\n\n## ⚙️ Configuration\nFor those using a non-Intility tenant, you also need to change make changes to the `provider_config`:\n\n```python\nfrom fastapi_azure_auth.provider_config import provider_config\n\nintility_scheme = AzureAuthorizationCodeBearer(\n    ...\n)\n\nprovider_config.tenant_id = \'my-own-tenant-id\'\n```\n\n\nIf you want, you can deny guest users to access your API by passing the `allow_guest_users=False`\nto `AzureAuthorizationCodeBearer`:\n\n```python\nintility_scheme = AzureAuthorizationCodeBearer(\n    ...\n    allow_guest_users=False\n)\n```\n',
    'author': 'Jonas Krüger Svensson',
    'author_email': 'jonas.svensson@intility.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/intility/fastapi-azure-auth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
